/**
 * Copyright (c) 2013 Saddle Development Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **/

package org.saddle.vec

import scala.{ specialized => spec }
import org.saddle._

// Specialized method implementations for code reuse in implementations of Vec; NA-safe
private[saddle] object VecImpl {
  def mask[@spec(Boolean, Int, Long, Double) A](
    v1: Vec[A], v2: Vec[Boolean], value: A)(implicit ma: CLM[A]): Vec[A] = {
    require(v1.length == v2.length, "Vectors must be the same length")
    val buf = Array.ofDim[A](v1.length)
    var i = 0
    while(i < v1.length) {
      val a = v1(i)
      val b = v2(i)
      buf(i) = if (!b) a else value
      i += 1
    }
    Vec[A](buf)
  }

  def mask[@spec(Boolean, Int, Long, Double) A: CLM](
    v1: Vec[A], f: A => Boolean, value: A): Vec[A] = {
    val sa = scalar.getScalarTag[A]
    val buf = Array.ofDim[A](v1.length)
    var i = 0
    while(i < v1.length) {
      val a = v1(i)
      buf(i) = if (sa.isMissing(a) || !f(a)) a else value
      i += 1
    }
    Vec[A](buf)
  }

  def foldLeft[@spec(Boolean, Int, Long, Double) A: CLM, @spec(Boolean, Int, Long, Double) B](
    vec: Vec[A])(init: B)(f: (B, A) => B): B = {
    val sa = scalar.getScalarTag[A]
    var acc = init
    var i = 0
    while(i < vec.length) {
      val v = vec(i)
      if (sa.notMissing(v)) acc = f(acc, v)
      i += 1
    }
    acc
  }

  /**
   * Same as foldLeft, but with a condition that operates on the accumulator and element
   * that if false, breaks out of the fold
   */
  def foldLeftWhile[@spec(Boolean, Int, Long, Double) A: CLM, @spec(Boolean, Int, Long, Double) B](
    vec: Vec[A])(init: B)(f: (B, A) => B)(cond: (B, A) => Boolean): B = {
    val sa = scalar.getScalarTag[A]
    var acc = init
    var i = 0
    while(i < vec.length) {
      val v = vec(i)
      if (sa.notMissing(v)) {
        if (cond(acc, v))
          acc = f(acc, v)
        else
          i = vec.length
      }
      i += 1
    }
    acc
  }

  def map[@spec(Boolean, Int, Long, Double) A: CLM, @spec(Boolean, Int, Long, Double) B: CLM](
    vec: Vec[A])(f: A => B): Vec[B] = {
    val sca = scalar.getScalarTag[A]
    val scb = scalar.getScalarTag[B]
    val buf = Array.ofDim[B](vec.length)
    var i = 0
    while(i < vec.length) {
      val v = vec(i)
      if (sca.isMissing(v))
        buf(i) = scb.missing
      else
        buf(i) = f(v)
      i += 1
    }
    Vec[B](buf)
  }

  /**
   * Save as foldLeft, but store and return intermediate accumulated results
   */
  def scanLeft[@spec(Boolean, Int, Long, Double) A: CLM, @spec(Boolean, Int, Long, Double) B: CLM](
    vec: Vec[A])(init: B)(f: (B, A) => B): Vec[B] = {
    val sca = scalar.getScalarTag[A]
    val scb = scalar.getScalarTag[B]
    val buf = Array.ofDim[B](vec.length)
    var acc = init
    var i = 0
    while(i < vec.length) {
      val v = vec(i)
      if (sca.notMissing(v)) {
        acc = f(acc, v)
        buf(i) = acc
      }
      else {
        buf(i) = scb.missing
      }
      i += 1
    }
    Vec[B](buf)
  }

  def zipMap[@spec(Int, Long, Double) A: CLM, @spec(Int, Long, Double) B: CLM,
             @spec(Boolean, Int, Long, Double) C: CLM](
    v1: Vec[A], v2: Vec[B])(f: (A, B) => C): Vec[C] = {
    require(v1.length == v2.length, "Vectors must be the same length")
    val sca = scalar.getScalarTag[A]
    val scb = scalar.getScalarTag[B]
    val scc = scalar.getScalarTag[C]
    val buf = Array.ofDim[C](v1.length)
    var i = 0
    while(i < v1.length) {
      val a = v1(i)
      val b = v2(i)
      if (sca.isMissing(a) || scb.isMissing(b)) {
        buf(i) = scc.missing
      }
      else {
        buf(i) = f(a, b)
      }
      i += 1
    }
    Vec[C](buf)
  }

  def filterFoldLeft[@spec(Boolean, Int, Long, Double) A: CLM, @spec(Boolean, Int, Long, Double) B](
    vec: Vec[A])(pred: (A) => Boolean)(init: B)(f: (B, A) => B): B = {
    val sa = scalar.getScalarTag[A]
    var acc = init
    var i = 0
    while(i < vec.length) {
      val vi = vec(i)
      if (sa.notMissing(vi) && pred(vi)) {
        acc = f(acc, vi)
      }
      i += 1
    }
    acc
  }

  def filterScanLeft[@spec(Boolean, Int, Long, Double) A: CLM, @spec(Boolean, Int, Long, Double) B: CLM](
    vec: Vec[A])(pred: (A) => Boolean)(init: B)(f: (B, A) => B): Vec[B] = {
    val sa = scalar.getScalarTag[A]
    val sb = scalar.getScalarTag[B]
    val buf = Array.ofDim[B](vec.length)
    var acc = init
    var i = 0
    while(i < vec.length) {
      val v = vec(i)
      if (sa.notMissing(v) && pred(v)) {
        acc = f(acc, v)
        buf(i) = acc
      }
      else {
        buf(i) = sb.missing
      }
      i += 1
    }
    Vec[B](buf)
  }

  def rolling[@spec(Boolean, Int, Long, Double) A, @spec(Boolean, Int, Long, Double) B: CLM](
    vec: Vec[A])(winSz: Int, f: Vec[A] => B): Vec[B] = {
    if (vec.length == 0 || winSz > vec.length)
      Vec.empty
    else {
      val buf = new Array[B](vec.length - winSz + 1)
      var i = winSz
      while (i <= vec.length) {
        buf(i - winSz) = f(vec.slice(i - winSz, i))
        i += 1
      }
      Vec(buf)
    }
  }

  def foreach[@spec(Boolean, Int, Long, Double) A: CLM](vec: Vec[A])(op: A => Unit) {
    val sa = scalar.getScalarTag[A]
    var i = 0
    while (i < vec.length) {
      val v: A = vec(i)
      if (sa.notMissing(v)) op(v)
      i += 1
    }
  }

  def forall[@spec(Boolean, Int, Long, Double) A: CLM](vec: Vec[A])(pred: A => Boolean)(op: A => Unit) {
    val sa = scalar.getScalarTag[A]
    var i = 0
    while (i < vec.length) {
      val v: A = vec(i)
      if (sa.notMissing(v) && pred(v)) op(v)
      i += 1
    }
  }

  def find[@spec(Boolean, Int, Long, Double) A: CLM](vec: Vec[A])(pred: A => Boolean): Vec[Int] = {
    val sa = scalar.getScalarTag[A]
    var i = 0
    val buf = Buffer[Int]()
    while(i < vec.length) {
      val v: A = vec(i)
      if (sa.notMissing(v) && pred(v)) buf.add(i)
      i += 1
    }
    Vec(buf.toArray)
  }

  def findOneNA[@spec(Boolean, Int, Long, Double) A: CLM](vec: Vec[A]): Boolean = {
    val sa = scalar.getScalarTag[A]
    var ex = false
    var i = 0
    while (!ex && i < vec.length) {
      val v: A = vec(i)
      ex = sa.isMissing(v)
      i += 1
    }
    ex
  }

  def isAllNA[@spec(Boolean, Int, Long, Double) A: CLM](vec: Vec[A]): Boolean = {
    val sa = scalar.getScalarTag[A]
    var ex = true
    var i = 0
    while (ex && i < vec.length) {
      val v: A = vec(i)
      ex = ex && sa.isMissing(v)
      i += 1
    }
    ex
  }

  def findOne[@spec(Boolean, Int, Long, Double) A: CLM](vec: Vec[A])(pred: A => Boolean): Int = {
    val sa = scalar.getScalarTag[A]
    var ex = false
    var i = 0
    while (!ex && i < vec.length) {
      val v: A = vec(i)
      ex = sa.notMissing(v) && pred(v)
      i += 1
    }
    if (ex) i - 1 else -1
  }

  def filter[@spec(Boolean, Int, Long, Double) A: CLM](vec: Vec[A])(pred: A => Boolean): Vec[A] = {
    val sa = scalar.getScalarTag[A]
    var i = 0
    val buf = Buffer[A]()
    while(i < vec.length) {
      val v: A = vec(i)
      if (sa.notMissing(v) && pred(v)) buf.add(v)
      i += 1
    }
    Vec(buf.toArray)
  }

  def where[@spec(Boolean, Int, Long, Double) A](vec: Vec[A])(pred: Array[Boolean])(
    implicit ev: CLM[A]): Vec[A] = {
    var i = 0
    val buf = Buffer[A]()
    while(i < vec.length) {
      val v: A = vec(i)
      if (pred(i)) buf.add(v)
      i += 1
    }
    Vec(buf.toArray)
  }

  def pad[@spec(Boolean, Int, Long, Double) A: CLM](vec: Vec[A], atMost: Int = 0): Vec[A] = {
    if (vec.length == 0 || vec.length == 1)
      vec
    else {
      val lim = if (atMost > 0) atMost else vec.length
      val sa = scalar.getScalarTag[A]
      val buf = array.empty[A](vec.length)
      buf(0) = vec(0)
      var i = 1
      var c = lim
      while(i < vec.length) {
        val v: A = vec(i)
        if (sa.notMissing(v)) {
          buf(i) = v
          c = lim
        }
        else {
          if (c > 0) buf(i) = buf(i - 1) else buf(i) = v
          c -= 1
        }
        i += 1
      }
      Vec(buf)
    }
  }

  def vecfillNA[@spec(Boolean, Int, Long, Double) A: CLM](vec: Vec[A])(f: (Int) => A): Vec[A] = {
    val buf = vec.contents
    var i = 0
    val l = vec.length
    val s = scalar.getScalarTag[A]
    while (i < l) {
      if (s.isMissing(buf(i))) buf(i) = f(i)
      i += 1
    }
    Vec(buf)
  }

  def seriesfillNA[@spec(Int, Long, Double) X, @spec(Boolean, Int, Long, Double) A: CLM](
    idx: Vec[X], vec: Vec[A])(f: X => A): Vec[A] = {
    val buf = vec.contents
    var i = 0
    val l = vec.length
    val s = scalar.getScalarTag[A]
    while (i < l) {
      if (s.isMissing(buf(i))) buf(i) = f(idx(i))
      i += 1
    }
    Vec(buf)
  }
}
